// Copyright (C) Microsoft Corporation. All rights reserved.

using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;
using System.IO;
using Microsoft.VSPowerToys.VSCmdShell;

namespace Microsoft.VSPowerToys.VSCmdShell.UnitTests
{
    /// <summary>
    /// Unit tests for cmd.exe host.
    /// </summary>
    [TestFixture]
    public class CommandShellHostTests
    {
        private IShellHost cmdHost;

        private MockUI shellUI = new MockUI();

        [SetUp]
        public void SetUpShell()
        {
            this.cmdHost = new Hosts.CommandShellHost();
            this.cmdHost.InitializeShellProcess(shellUI, null);
        }

        [TearDown]
        public void DisposeShell()
        {
            if (this.cmdHost != null)
            {
                this.cmdHost.Dispose();
            }
        }

        /// <summary>
        /// This test checks if execute command works correctly for the cmd.exe host.
        /// 
        /// The way test works is, we first get a temporary filename and remove the created file. Then
        /// we "issue mkdir temppath" command to create a directory using the filename we got. 
        /// If shell host is working correctly, a directory should be created. 
        /// 
        /// We issue a rmdir command at the end to remove the directory.
        /// </summary>
        [Test]
        public void CommandIssueTest()
        {
            string tempPath = Path.GetTempFileName();
            File.Delete(tempPath);
            this.cmdHost.ExecuteCommand("mkdir \"" + tempPath + "\"");
            /// Sleeping to allow command to execute
            System.Threading.Thread.Sleep(1000);
            Assert.IsTrue(Directory.Exists(tempPath), tempPath);
            this.cmdHost.ExecuteCommand("rmdir \"" + tempPath + "\"");
            System.Threading.Thread.Sleep(1000);
            Assert.IsFalse(Directory.Exists(tempPath));
        }

        /// <summary>
        /// This test checks if echo command correctly returns echoed text
        /// </summary>
        [Test]
        public void EchoTest()
        {
            System.Threading.Thread.Sleep(500);
            this.shellUI.ResetOutput();
            //Verify that international characters work properly
            this.cmdHost.ExecuteCommand("echo TstString");
            /// Sleeping to allow command to execute
            System.Threading.Thread.Sleep(1000);
            Assert.AreEqual("TstString", this.shellUI.Output.ToString().Substring(0, "TstString".Length), "Output did not contain expected string");
        }

        /// <summary>
        /// This test checks if directory parameter is correctly handled by CommandShellHost
        /// </summary>
        [Test]
        public void StartingDirectoryTest()
        {
            if (this.cmdHost != null)
            {
                this.cmdHost.Dispose();
            }
            this.cmdHost = new Hosts.CommandShellHost();
            this.cmdHost.InitializeShellProcess(shellUI, Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles));
            System.Threading.Thread.Sleep(500);
            this.shellUI.ResetOutput();
            this.cmdHost.ExecuteCommand("cd");
            /// Sleeping to allow command to execute
            System.Threading.Thread.Sleep(1000);
            Assert.IsTrue(
                this.shellUI.Output.ToString().StartsWith(Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles)),
                "Output did not contain expected string");
        }
    }

    /// <summary>
    /// Mock shell UI for testing purposes
    /// </summary>
    class MockUI : IShellUI
    {

        private StringBuilder output;

        /// <summary>
        /// Output from shell host process.
        /// </summary>
        public StringBuilder Output
        {
            get
            {
                if (output == null)
                {
                    output = new StringBuilder();
                }
                return output;
            }
        }

        /// <summary>
        /// Clears the output from host process
        /// </summary>
        public void ResetOutput()
        {
            this.output = null;
        }

        #region IShellUI Members

        public bool ApplicationRunning
        {
            get
            {
                return true;
            }
            set
            {
            }
        }

        public int WindowHeight
        {
            get { return 80; }
        }

        public int WindowWidth
        {
            get { return 40; }
        }

        public void WriteLine(string value)
        {
            this.Output.AppendLine(value);
        }

        public void Write(string value)
        {
            this.Output.Append(value);
        }

        public string ReadLine()
        {
            return String.Empty;
        }

        public event System.Windows.Forms.KeyEventHandler KeyDown;

        public event System.Windows.Forms.KeyEventHandler KeyUp;

        #endregion

        /// <summary>
        /// Placeholder method
        /// </summary>
        public void InvokeEvents()
        {
            if (this.KeyDown != null)
            {
                this.KeyDown.Invoke(this, null);
            }
            if (this.KeyUp != null)
            {
                this.KeyUp.Invoke(this, null);
            }

        }
    }
}